#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include "pktlib.h"
#include "pktbuf.h"

/*
 * ʲܻ2014ǯýֽθ饺٥꡼ѥܳʥͥåȥ
 * 裱Ǻʰץʥ饤(analyze.c)١ˡѥåȥХåե
 * Ѥ褦˲¤Ρ
 * http://kozos.jp/books/interface/201408/recv-packet/analyzer/
 */

static void proc_arp(pktbuf_t pktbuf)
{
  struct arphdr *arphdr;
  struct in_addr ipaddr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  arphdr = (struct arphdr *)p;
  printf("ARP\toperation: %d\n", ntohs(arphdr->ar_op));
  p += sizeof(*arphdr);
  printf("\tsource MAC:%s\n", ether_ntoa((struct ether_addr *)p));
  p += arphdr->ar_hln; memcpy(&ipaddr, p, sizeof(ipaddr));
  printf("\tsource IP :%s\n", inet_ntoa(ipaddr));
  p += arphdr->ar_pln;
  printf("\ttarget MAC:%s\n", ether_ntoa((struct ether_addr *)p));
  p += arphdr->ar_hln; memcpy(&ipaddr, p, sizeof(ipaddr));
  printf("\ttarget IP :%s\n", inet_ntoa(ipaddr));
}

static void proc_icmp(pktbuf_t pktbuf)
{
  struct icmp *icmphdr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  icmphdr = (struct icmp *)p;
  printf("ICMP\ttype: %d\n", icmphdr->icmp_type);
  printf("\tcode: %d\n", icmphdr->icmp_code);
}

static void proc_tcp(pktbuf_t pktbuf)
{
  struct tcphdr *tcphdr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  tcphdr = (struct tcphdr *)p;
  printf("TCP\tsrc port: %d\n", ntohs(tcphdr->th_sport));
  printf("\tdst port: %d\n", ntohs(tcphdr->th_dport));
  printf("\tseq number: %u\n", ntohl(tcphdr->th_seq));
  printf("\tack number: %u\n", ntohl(tcphdr->th_ack));
  printf("\tflags: 0x%02x\n", tcphdr->th_flags);
}

static void proc_udp(pktbuf_t pktbuf)
{
  struct udphdr *udphdr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  udphdr = (struct udphdr *)p;
  printf("UDP\tsrc port: %d\n", ntohs(udphdr->uh_sport));
  printf("\tdst port: %d\n", ntohs(udphdr->uh_dport));
  printf("\tlength : %d\n", ntohs(udphdr->uh_ulen));
}

static void proc_ip(pktbuf_t pktbuf)
{
  struct ip *iphdr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  iphdr = (struct ip *)p;
  printf("IP\theader size: %d bytes\n", (iphdr->ip_hl << 2));
  printf("\ttotal size: %d bytes\n", ntohs(iphdr->ip_len));
  printf("\tTTL: %d\n", iphdr->ip_ttl);
  printf("\tprotocol: %d\n", iphdr->ip_p);
  printf("\tsrc IP addr: %s\n", inet_ntoa(iphdr->ip_src));
  printf("\tdst IP addr: %s\n", inet_ntoa(iphdr->ip_dst));

  pktbuf_delete_header(pktbuf, iphdr->ip_hl << 2); /* IPإå */
  switch (iphdr->ip_p) { /* ̥ץȥ˱ƳƼƤӽФ */
  case IPPROTO_ICMP: proc_icmp(pktbuf); break;
  case IPPROTO_TCP:  proc_tcp( pktbuf); break;
  case IPPROTO_UDP:  proc_udp( pktbuf); break;
  default: break;
  }
}

#ifdef __FreeBSD__
static void proc_loopback(pktbuf_t pktbuf)
{
  u_int32_t family;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  family = *(int *)p; /* ntohl() */
  printf("family: %d\n", family);
  pktbuf_delete_header(pktbuf, sizeof(family)); /* loopbackإå */
  switch (family) {
  case AF_INET: proc_ip(pktbuf); break;
  default: break;
  }
}
#endif

static void proc_ethernet(pktbuf_t pktbuf)
{
  struct ether_header *ehdr;
  char *p = pktbuf_get_header(pktbuf); /* ץΰ褫إå֤ */
  ehdr = (struct ether_header *)p;
  printf("%s ->", ether_ntoa((struct ether_addr *)ehdr->ether_shost));
  printf(" %s  ", ether_ntoa((struct ether_addr *)ehdr->ether_dhost));
  printf("(type: 0x%04x)\n", ntohs(ehdr->ether_type));
  pktbuf_delete_header(pktbuf, ETHER_HDR_LEN); /* Ethernetإå */
  switch (ntohs(ehdr->ether_type)) {
  case ETHERTYPE_ARP: proc_arp(pktbuf); break;
  case ETHERTYPE_IP:  proc_ip( pktbuf); break;
  default: break;
  }
}

int main(int argc, char *argv[])
{
  pktif_t pktif;
  pktbuf_t pktbuf;
  int linktype, size;
  char *p;

  pktif = pktif_open(argv[1], 0, 0); /* 󥿡ե򥪡ץ󤹤 */

  linktype = pktif_get_linktype(pktif);
  printf("linktype: %d\n", linktype);

  while (1) {
    pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
    if (pktbuf == NULL) continue;

    p = pktbuf_get_header(pktbuf); /* ץΰ褫ХåեƬ */
    size = pktbuf_get_size(pktbuf); /* ץΰ褫饵 */
    printf("----\nreceived: %d bytes (buffer:%08x)\n", size, (int)p);

    switch (linktype) {
#ifdef __FreeBSD__
    case DLT_NULL: proc_loopback(pktbuf); break; /* ѥåȤβϽ */
#endif
#ifdef __linux__
    case DLT_NULL: /* fall through */
#endif
    case DLT_EN10MB: proc_ethernet(pktbuf); break; /* ѥåȤβϽ */
    default: break;
    }

    pktbuf_destroy(pktbuf);
    fflush(stdout);
  }

  return 0;
}
